// Portable implementation of the bipartite table method for a signaling
// less-than-or-equal comparison between a binary64 and a decimal64
// 
// Copyright (C) 2012, 2013 by
// 
// Laboratoire de l'Informatique du Parallelisme,
// UMR CNRS - ENS Lyon - UCB Lyon 1 - INRIA 5668
// 
// and by
// 
// Laboratoire d'Informatique de Paris 6, equipe PEQUAN,
// UPMC Universite Paris 06 - CNRS - UMR 7606 - LIP6, Paris, France.
// 
// Authors : Christoph Lauter, Marc Mezzarobba
// 
// Contact : christoph.lauter@lip6.fr, marc@mezzarobba.net
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
// 

#include <stdio.h>
#include <stdint.h>

/* Get our constants and types */
#include "comparison.h"
#include <math.h>

/* Get the table we need for the comparison */
#include "comparison-table-alt.h"

/* Compile me with something like 

    icc -xSSE4.2 -O3 -c -o comparison-alt.o comparison-alt.c 

*/

#define multiplyUint64ByUint64(multiplyUint64ByUint64_rh,                                                         \
                               multiplyUint64ByUint64_rl,                                                         \
                               multiplyUint64ByUint64_a,                                                          \
                               multiplyUint64ByUint64_b) do {                                                     \
  uint64_t multiplyUint64ByUint64_ah;                                                                             \
  uint64_t multiplyUint64ByUint64_al;                                                                             \
  uint64_t multiplyUint64ByUint64_bh;                                                                             \
  uint64_t multiplyUint64ByUint64_bl;                                                                             \
  uint64_t multiplyUint64ByUint64_resl;                                                                           \
  uint64_t multiplyUint64ByUint64_resm1;                                                                          \
  uint64_t multiplyUint64ByUint64_resm2;			                                                  \
  uint64_t multiplyUint64ByUint64_resm;                                                                           \
  uint64_t multiplyUint64ByUint64_resmh;                                                                          \
  uint64_t multiplyUint64ByUint64_resml;                                                                          \
  uint64_t multiplyUint64ByUint64_resh;	                                                                          \
  multiplyUint64ByUint64_ah = (multiplyUint64ByUint64_a) >> 32;                                                   \
  multiplyUint64ByUint64_al = (multiplyUint64ByUint64_a) & 0x00000000ffffffffull;                                 \
  multiplyUint64ByUint64_bh = (multiplyUint64ByUint64_b) >> 32;                                                   \
  multiplyUint64ByUint64_bl = (multiplyUint64ByUint64_b) & 0x00000000ffffffffull;                                 \
  multiplyUint64ByUint64_resl = multiplyUint64ByUint64_al * multiplyUint64ByUint64_bl;                            \
  multiplyUint64ByUint64_resh = multiplyUint64ByUint64_ah * multiplyUint64ByUint64_bh;                            \
  multiplyUint64ByUint64_resm1 = multiplyUint64ByUint64_ah * multiplyUint64ByUint64_bl;                           \
  multiplyUint64ByUint64_resm2 = multiplyUint64ByUint64_al * multiplyUint64ByUint64_bh;                           \
  multiplyUint64ByUint64_resm = multiplyUint64ByUint64_resm1 + multiplyUint64ByUint64_resm2;                      \
  multiplyUint64ByUint64_resh += ((uint64_t) (multiplyUint64ByUint64_resm < multiplyUint64ByUint64_resm1)) << 32; \
  multiplyUint64ByUint64_resmh = multiplyUint64ByUint64_resm >> 32;                                               \
  multiplyUint64ByUint64_resml = multiplyUint64ByUint64_resm << 32;                                               \
  *(multiplyUint64ByUint64_rl) = multiplyUint64ByUint64_resl + multiplyUint64ByUint64_resml;                      \
  multiplyUint64ByUint64_resh += (uint64_t) (*(multiplyUint64ByUint64_rl) < multiplyUint64ByUint64_resl);         \
  *(multiplyUint64ByUint64_rh) = multiplyUint64ByUint64_resh + multiplyUint64ByUint64_resmh;                      \
} while (0) 

#define multiplyTwoUint64ByUint64(multiplyTwoUint64ByUint64_rh,                                                   \
                                  multiplyTwoUint64ByUint64_rm,                                                   \
                                  multiplyTwoUint64ByUint64_rl,                                                   \
                                  multiplyTwoUint64ByUint64_ah,                                                   \
                                  multiplyTwoUint64ByUint64_al,                                                   \
                                  multiplyTwoUint64ByUint64_b) do {                                               \
  uint64_t multiplyTwoUint64ByUint64_t, multiplyTwoUint64ByUint64_c;                                              \
  multiplyUint64ByUint64(&multiplyTwoUint64ByUint64_c, (multiplyTwoUint64ByUint64_rl),                            \
                         (multiplyTwoUint64ByUint64_al), (multiplyTwoUint64ByUint64_b));                          \
  multiplyUint64ByUint64((multiplyTwoUint64ByUint64_rh), &multiplyTwoUint64ByUint64_t,                            \
                         (multiplyTwoUint64ByUint64_ah), (multiplyTwoUint64ByUint64_b));                          \
  *(multiplyTwoUint64ByUint64_rm) = multiplyTwoUint64ByUint64_c + multiplyTwoUint64ByUint64_t;                    \
  *(multiplyTwoUint64ByUint64_rh) += (uint64_t) (*(multiplyTwoUint64ByUint64_rm) < multiplyTwoUint64ByUint64_c);  \
} while (0)

int compareSignalingLessEqualBinary64Decimal64(_binary64 x, _Decimal64 y) {
  bin64caster xcst;
  dec64caster ycst;
  volatile _binary64 a, b, c;
  int xIsInf, yIsInf, xIsZero, yIsZero, xIsNeg, yIsNeg;
  uint64_t ySignificand;
  int64_t E, F, G, S, H, cTwo19Log2OverLog5, Q, q, r, cTwo12Log5OverLog2, sigma;
  uint64_t m, n, t, theta1hi, theta1lo, theta2, lefth, leftl, righth, rightm, rightl;
  uint64_t resh, resl; 
  int64_t reshsigned;
  bin64caster temp;

  xcst.f = x;
  ycst.f = y;

  /* Handle NaNs first: for a signaling comparison for less-or-equal
     the result always is false and invalid is to be signaled. 
  */
  if (((xcst.i & 0x7fffffffffffffffull) > 0x7ff0000000000000ull) ||
      ((ycst.i & 0x7c00000000000000ull) == 0x7c00000000000000ull)) {
    /* At least one of x or y is NaN.
       Perform two dummy operations and return false.
    */
    a = 0.0;
    b = 0.0;
    c = a / b; /* Signals invalid */
    return 0;
  }

  /* Here, neither x nor y is NaN */

  /* Check if x or y are zero. Also determine their signs. */
  xIsNeg = ((xcst.i & 0x8000000000000000ull) != 0x0ull);
  yIsNeg = ((ycst.i & 0x8000000000000000ull) != 0x0ull);
  xIsZero = ((xcst.i & 0x7fffffffffffffffull) == 0x0ull);
  ySignificand = (((ycst.i & 0x6000000000000000ull) != 0x6000000000000000ull) ? 
		  (ycst.i & 0x001fffffffffffffull) : 
		  ((ycst.i & 0x0007ffffffffffffull) | 0x0020000000000000ull));
  if (ySignificand > 9999999999999999ull) ySignificand = 0x0ull;
  yIsZero = (ySignificand == 0x0ull);
  
  if (xIsZero || yIsZero) {
    /* Here, at least one of x or y is zero. */

    /* Start by checking if x and y are both zero */
    if (xIsZero && yIsZero) {
      /* Here, both x and y are zero. As we for a -or-equal comparison
	 the result is true, regardless of the sign of the zeros. 
      */
      return 1;
    }

    /* Here, only one of x or y is zero */
    if (xIsZero) {
      /* Here, x is zero and y is non-zero. The sign of y determines
	 the result: for a less-or-equal comparison, the result is
	 true iff y is not negative.
      */
      return !yIsNeg;
    }

    /* Here, y is zero and x is non-zero. The sign of x determines the
       result: for a less-or-equal comparison, the result is true iff
       x is negative.
    */
    return xIsNeg;
  }

  /* Here, x and y are both neither zero nor NaN. */
  
  /* Now perform a first check based on the sign of the input: 
     if the signs of x and y are different, the result is already known.
  */
  if (xIsNeg != yIsNeg) {
    /* Here, the signs of x and y are different.  The result is
       already known: the less-or-equal comparison is true iff x is
       negative (and y is positive).
    */
    return xIsNeg;
  }
  
  /* Here, x and y are both neither zero nor NaN and have the same
     sign. 
  */
  
  /* Now check if x or y is +/- Infinity */
  xIsInf = ((xcst.i & 0x7fffffffffffffffull) == 0x7ff0000000000000ull);
  yIsInf = ((ycst.i & 0x7c00000000000000ull) == 0x7800000000000000ull);
  if (xIsInf || yIsInf) {
    /* Here, at least one of x or y is infinity. We do know that they
       have the same signs. So we start by testing if they are both
       infinities.
    */
    if (xIsInf && yIsInf) {
      /* Here, both x and y are infinity. For a less-or-equal
	 comparison, the result is hence true.
      */
      return 1;
    }
    
    /* Here, at least one of x and y is a finite number, the other is
       an infinity.
    */
    if (xIsInf) {
      /* Here x is an infinity, y is finite. The result of the
	 less-than-equal comparison is true iff x is negative. 
      */
      return xIsNeg;
    }

    /* Here x is finite and non-zero, y is an infinity. The result of
       the less-than-equal comparison is true iff y is not negative.
    */
    return !yIsNeg;
  }

  /* Here, x and y are both non-zero finite numbers of the same
     sign. 
  */

  /* We start by decomposing |x| into 

     2^E * m = |x|

     with E integer and m integer satisfying 2^52 <= m <= 2^53 - 1.

  */
  xcst.i &= 0x7fffffffffffffffull;
  E = 0;
  if (xcst.i < 0x0010000000000000ull) {
    /* Handle subnormal numbers */
    E -= 54;
    xcst.f *= 18014398509481984.0; /* Multiply with 2^54 */
  }
  E += ((xcst.i & 0x7ff0000000000000ull) >> 52) - 1023 - 52;
  m = (xcst.i & 0x000fffffffffffffull) | 0x0010000000000000ull;
  
  /* Now we decompose |y| into 

     2^F * 5^G * n = |y|

     with F integer and n integer satisfying 2^53 <= m <= 2^54 - 1.

     We already know the binary encoded significand ySignificand that
     satisfies:

     1 <= ySignificand <= 10^16 - 1.

     We have 10^16 - 1 \approx 2^53.151 hence 2^53 < 10^16 - 1 < 2^54.

     We first compute G such that 

     10^G * ySignificand = |y|

     Then we use the binary FP unit as a leading zero counter in order
     to normalize ySignificand for F and n, where 2^53 <= n <=
     2^54 - 1.

  */
  G = (((ycst.i & 0x6000000000000000ull) != 0x6000000000000000ull) ? 
       ((ycst.i & 0x7fe0000000000000ull) >> 53) : 
       ((ycst.i & 0x1ff8000000000000ull) >> 51)) - 398;

  t = ySignificand >> 32;
  S = 0;
  if (t == 0ull) {
    t = ySignificand;
    S = 32;
  } 
  temp.i = 0x4330000000000000 | t; /* 0x43300...0 = 2^52 */
  temp.f -= 4503599627370496.0;    /* 2^52 */
  S += (53 - 32) - ((temp.i >> 52) - 1023);
  n = ySignificand << S;
  F = G - S;

  /* Now we have to compare 2^E * m with 2^F * 5^G * n with 

     E, F, G integers

     and 

     2^52 <= m <= 2^53 - 1

     as well as

     2^53 <= n <= 2^54 - 1.

     This is equivalent to comparing 

     2^(E - F) * m with 5^G * n.

     Let H be H = E - F.
  */
  H = E - F;

  /* We can eliminate quite a few cases considering only the
     "exponents" G and H due to the following two implications we can
     prove:

     For m and n normalized as indicated, we have

     (1) G > floor(H * log(2)/log(5)) => 5^G * n > 2^H * m
     (2) G < floor(H * log(2)/log(5)) => 5^G * n < 2^H * m

     So we have to compute the quantity 

     Q = floor(H * log(2)/log(5))

     We can show that H is an integer in the range 

     -1495 <= H <= 1437

     and that for all these integers, we have

     floor(H * log(2)/log(5)) = floor(2^(-19) * H * 
                                     nearestint(2^19 * log(2)/log(5))).
     
     Let be

     cTwo19Log2OverLog5 = nearestint(2^19 * log(2)/log(5)) = 225799.

     So we can compute Q by a simple multiplication followed by a shift.

     It is easy to show that there is no integer overflow for at least
     signed 32 bit integers.

  */
  cTwo19Log2OverLog5 = 225799;
  Q = (H * cTwo19Log2OverLog5) >> 19;

  /* Now check if G < Q or G > Q or if we have equality */
  if (Q != G) {
    /* Here we have either Q < G or Q > G */
    if (Q < G) {
      /* Here we have Q < G hence 5^G * n > 2^H * m.

	 So we know that 

	 abs(x) < abs(y)   and sgn(x) = sgn(y)

	 which means that the less-or-equal comparison x <= y 
	 is true iff x and y are not negative.

      */
      return !xIsNeg;
    } 
    
    /* Here we have Q > G hence 5^G * n < 2^H * m 

       So we know that 

       abs(x) > abs(y)   and sgn(x) = sgn(y)

       which means that the less-or-equal comparison x <= y 
       is true iff x and y are negative.
       
    */
    return xIsNeg;
  } 

  /* If we are here, we have G = Q and we must consider the
     mantissas. 

     The comparison problem is to compare

        2^H * m to 5^G * n

     As we have G = Q = floor(H * log(2)/log(5)), this is 
     the same as comparing

        m to 5^G * 2^(-H) * n

     which is in turn equivalent to comparing

        theta_2(r) * 2^sigma * 2^8 * m to theta_1(q) * 2^-64 * 2^8 * n

     where

        q = floor(G/16) + 1
        r = 16 * q - G

        theta_1(q) = 5^(16 * q) * 2^127 * 2^(-floor(16 * q * log(5)/log(2)))
        theta_2(r) = 5^r * 2^63 * 2^(-floor(r * log(5)/log(2)))

        sigma = floor(r * log(5)/log(2)) - floor(16 * q * log(5)/log(2)) + H.

      We start by computing q, r and sigma.

      As G varies between -413 and 369, q varies between -25 and 24.
      It is clear that r varies between 1 and 16.

      It can be shown that sigma is a function of H, whose range is
      bounded by -958 and 859, and that sigma varies between 0 and 3.

      It can also be shown that for q integer in the range -25 and 24, we have
      
       floor(16 * q * log(5)/log(2)) = 
          = floor(2^-8 * q * nearestint(16 * 2^8 * log(5)/log(2)))

      where

       nearestint(16 * 2^8 * log(5)/log(2)) = 9511.

      It can finally be shown that this same constant can be used again in
  
       floor(r * log(5)/log(2)) = 
          = floor(2^-12 * r * nearestint(16 * 2^8 * log(5)/log(2)))

      for r integer between 1 and 16.

  */
  q = (G >> 4) + 1ll;
  r = (q << 4) - G;
  cTwo12Log5OverLog2 = 9511ll;
  sigma = ((r * cTwo12Log5OverLog2) >> 12) + H - ((q * cTwo12Log5OverLog2) >> 8);

  /* Now we compute m' = 2^(8 + sigma) * m and n' = 2^8 * n */
  m <<= 8 + sigma;
  n <<= 8;

  /* The comparison problem is now reduced to comparing

      theta_2(r) * m' to theta_1(q) * 2^-64 * n'.

     It can be shown that theta_2(r) is an integer in the range 

      0 < theta_2(r) <= 2^64 - 1 

     and can therefore be stored in a table exactly (without any
     approximation error).

     In contrast, theta_1(q) gets replaced by 

      theta_1'(q) = ceil(theta_1(q)) = theta_1(q) + deltaTbl

     with 0 <= deltaTbl < 1.

     It can be shown that we have

      theta_1'(q) = theta_1(q) * (1 + epsTbl)

     with 

      0 <= epsTbl < 2^(-127.21).

     We further have 0 <= theta_1'(q) <= 2^128 - 1, i.e. theta_1'(q)
     holds on two 64 bit unsigned integers.

     Additionnally, for q in the range 0 <= q <= 1, 
     we have 

      2^64 * floor(theta_1'(q) * 2^-64) = theta_1(q) 

     exactly.

     The same way, for q = 2, we have 

      2^53 * floor(theta_1'(q) * 2^-53) = theta_1(q).

     We start by loading theta1'(q) and theta(r) from three tables. We
     then exactly compute the products

      2^64 * lefth + leftl = theta2(r) * m'

     and 

      2^128 * righth + 2^64 * rightm + rightl = theta1'(q) * n'.

  */
  theta1hi = comparisonTblTheta1Hi[q - qmin];
  theta1lo = comparisonTblTheta1Lo[q - qmin];
  theta2   = comparisonTblTheta2[r - rmin];

  multiplyUint64ByUint64(&lefth, &leftl, theta2, m);
  multiplyTwoUint64ByUint64(&righth, &rightm, &rightl, theta1hi, theta1lo, n);

  /* As all multiplications are exact and as the table for theta_1(q)
     is more accurate than the worst-case for the associated
     approximation problem m/n approximated by 5^G/2^H (with the given
     bounds for m, n, G and H), the comparison problem is now reduced
     to comparing

      2^64 * lefth + leftl to 2^64 * righth + rightm + 2^-64 * rightl

     taking the following properties of the approximation into account:
     
     (i)   We have 

            2^128 * righth + 2^64 * rightm + rightl >= theta1(q) * n'.

	   Therefore, if 

	    2^64 * lefth + leftl > 2^64 * righth + rightm + 2^-64 * rightl

           then immediately

            2^64 * lefth + leftl > 2^-64 * theta1(q) * n'

           which implies

            2^H * m > 5^G * n.

           Further, if 

            2^64 * righth + rightm - 2^64 * lefth - leftl < 0

           then 

            2^64 * righth + rightm - 2^64 * lefth - leftl <= -1

           because righth, rightm, lefth and leftl are integers.

           Therefore, if 

            2^64 * righth + rightm - 2^64 * lefth - leftl < 0

           then 

            2^64 * righth + rightm + 2^-64 * rightl - 2^64 * lefth - leftl < 0,

           because 2^-1 <= 2^-64 * rightl <= 1 - 2^-64 < 1.

           Hence, if 

            2^64 * righth + rightm - 2^64 * lefth - leftl < 0

           then 

            2^64 * lefth + leftl > 2^64 * righth + rightm + 2^-64 * rightl.

           So, finally, if 

            2^64 * righth + rightm - 2^64 * lefth - leftl < 0

           then 

	    2^H * m > 5^G * n.

     (ii)  Since we have 

            G = 16 * q - r

           and since the high-order word of the table theta_1'(q) is
	   an exact representation of 2^-64 * theta_1(q) for q in the
	   range 0 <= q <= 1, we know that the "approximated"
	   comparison problem (comparing 2^64 * lefth + leftl and 2^64
	   * righth + rightm + 2^-64 * rightl) is exactly the
	   equivalent to the comparison between m and 5^G * 2^(-H) * n
	   for G in the range:

	    -16 <= G <= 15.

	   As we have G = floor(H * log(2)/log(5)), we have the same 
	   equivalence for H in the range 

             -37 <= H <= 37.

           Further, since for 0 <= q <= 1 the high-order word of 
	   the theta_1 table is already exact (modulo a scaling), the 
	   low-order word is necessarily zero:

            theta_1hi(q) = floor(2^-64 * theta_1(q))
            theta_1lo(q) = floor(theta_1(q) - 2^64 * theta_1hi(q))

           and 

            theta_1hi(q) = theta_1(q) for 0 <= q <= 1.

           The low-order word of theta_1'(q) being zero, the low-order
           word rightl of the product 2^128 * righth + 2^64 * rightm +
           rightl = theta_1'(q) * n' is zero, too.

           In consequence, the "approximated" comparison problem
	   neglecting the low-order word rightl of the product --
	   'compare 2^64 * lefth + leftl to 2^64 * righth + rightm' --
	   is exactly equivalent to the comparison between m and 5^G *
	   2^(-H) * n for G in the range -16 <= G <= 15, respectively
	   H in the range -37 <= H <= 37.

	   Similarily, if q = 2, we have 

	    2^53 * floor(theta_1'(q) * 2^-53) = theta_1(q).

           This means that if n is a multiple of 8, n' = 2^8 * n is a
           multiple of 2^11 and n' * theta_1(q) is a multiple of
           2^64. In consequence, in this case, the low-order word
           rightl of the product 2^128 * righth + 2^64 * rightm +
           rightl = theta_1'(q) * n' = theta_1(q) is zero. Hence, for
           q = 2 and n a multiple of 8, the "approximated" comparison
           problem neglecting the low-order word rightl of the product
           -- 'compare 2^64 * lefth + leftl to 2^64 * righth + rightm'
           -- is exactly equivalent to the comparison between m and
           5^G * 2^(-H) * n. The equality q = 2 implies that G is in
           the range 2 * 16 - 16 <= G <= 2 * 16 - 1, i.e. 16 <= G <=
           31. This implies that H is in the range 38 <= H <= 74.

           In conclusion: The "approximated" comparison problem

            compare 2^64 * lefth + leftl to 2^64 * righth + rightm

           is exactly equivalent to the comparison 

            compare m to 5^G * 2^(-H) * n

           a) if -37 <= H <= 37
           b) if  38 <= H <= 74 and n is a multiple of 8.

     (iii) For the case when q = -1, i.e. when -32 <= G <= -17,
           respectively when -74 <= H <= -38, we have the following 
           rounding-effect:

           Let n be a multiple of 5^17 in this case, i.e. let be 

            n = 5^17 * t.

           As n is bounded by 2^53 <= n <= 2^54 - 1, this means 
           that t is bounded by 11806 <= t <= 23611.

	   Notice that we had n' = 2^8 * n.

	   Since we have 

            theta_1'(q) = ceil(5^(16 * q) * 2^127 
                                 * 2^(-floor(16 * q * log(5)/log(2))))
           
           we have 

            theta_1'(-1) = 5^(-16) * 2^165 + delta = theta_1(-1) + delta.

           It can be shown that delta satisfies

            2^(-2) <= delta <= 2^(-2) + 2^(-8).

           Therefore theta_1'(-1) * n' satisfies

            theta_1'(-1) * n' = 2^173 * 5 * t + delta'

           with delta' = 2^8 * 5^17 * t * delta.

           It is easy to show that delta' satisfies

            2^59 <= delta' <= 2^61.

           In consequence we get in this case:

           2^128 * righth + 2^64 * rightm = 
             = 2^64 * floor(2^-64 * theta_1'(-1) * n')
             = 2^64 * floor(2^109 * 5 * t + delta'')

           with delta'' = 2^-64 * delta', satisfying

            0 < 2^-5 <= delta'' <= 2^-3 < 1

           It is clear that 2^109 * 5 * t is integer.

           Therefore we get

           2^128 * righth + 2^64 * rightm = 
             = 2^64 * floor(2^109 * 5 * t + delta'')
             = 2^173 * 5 * t 
             = theta_1(-1) * n'.

           This means for -74 <= H <= -38, respectively 
           -32 <= G <= -17 and n a multiple of 5^17, the
	   two high-order words 2^128 * righth + 2^64 * rightm
	   are exactly equal to theta_1(q) * n'.

	   This means that for -74 <= H <= -38, respectively -32 <= G
           <= -17 and n a multiple of 5^17, the "approximated"
           comparison problem neglecting the low-order word rightl of
           the product -- 'compare 2^64 * lefth + leftl to 2^64 *
           righth + rightm' -- is exactly equivalent to the comparison
           between m and 5^G * 2^(-H) * n.

     (iv) The following properties can be shown for the exact cases
          (equality cases) of the comparison m compared to 5^G *
          2^(-H) * n with the known bounds on m, n, G and H:

            a) All exact cases imply that -52 <= H <= 52.

            b) For 38 <= H <= 52, n is a multiple of 2^38, hence also 
               a multiple of 8.

            c) For -52 <= H <= -38, we have -23 <= G <= -17 and 
               n is a multiple of 5^17.

          These three properties, combined with points (ii) and (iii) 
          above, yield to the following implication:

          If m = 5^G * 2^(-H) * n holds (with the known bounds on m,
          n, G and H), then

	   2^64 * lefth + leftl = 2^64 * righth + rightm holds.

     (v)  We have 

           2^64 * righth + rightm = 
             = floor(2^-64 * theta_1'(q) * n')
             = floor(2^-64 * ceil(theta_1(q)) * n')
             = 2^-64 * theta_1(q) * n' + 2^-64 * deltaTbl * n' + deltaApprox
             = 2^-64 * theta_1(q) * n' + deltaAll 

           with 0 <= deltaTbl < 1 and -1 < deltaApprox <= 0.

           As 2^61 <= n' < 2^62, we get 

            -1 < deltaAll < 2^-64 * 1 * 2^62

           with means

            -1 < deltaAll < 2^-2 or abs(deltaAll) < 1

           We also have

            2^64 * theta_1(q) * n' = 
             = 2^63 * 5^(16 * q) * 2^(-16 * q * log(5)/log(2) - deltaStar) * n'
             = 2^63 * 2^(-deltaStar) * n'

           with -1 < deltaStar <= 0.

           Hence 

            2^63 * 1 * 2^61 <= 2^-64 * theta_1(q) * n' < 2^63 * 2 * 2^62

           which simplifies to 

            2^125 <= 2^-64 * theta_1(q) * n' < 2^126.

           Therefore 

            2^64 * righth + rightm = 2^-64 * theta_1(q) * n' * (1 + epsAll)

           with epsAll = deltaAll / (2^-64 * theta_1(q) * n') bounded by

            abs(epsAll) < 1 * 2^-124.

           We further repeat that we have 

            2^64 * lefth + leftl = theta_2(r) * 2^sigma * 2^8 * m 

           and that

            (5^G * n)/(2^H * m) = (2^-64 * theta_1(q) * n')/(theta_2(r) * 2^sigma * 2^8 * m).

	   This means that we have 

            (2^64 * righth + rightm)/(2^64 * lefth + leftl) = (5^G * n)/(2^H * m) * (1 + epsAll).

     (vi) We have G = floor(H * log(2)/log(5)) = H * log(2)/log(5) + deltaSpecial 
          with -1 < deltaSpecial <= 0.

          Therefore 5^G / 2^H satisfies

	   5^G / 2^H = 5^(H * log(2)/log(5) + deltaSpecial) / 2^H = 1 * 5^deltaSpecial.

          Hence 

           1/5 < 5^G / 2^H <= 1.

          Suppose now that 2^64 * righth + rightm = 2^64 * lefth + leftl but that 
          5^G * n != 2^H * m, i.e. that the case is inexact.

	  With the last equation given for point (v), this implies that 

           1 = (5^G * n)/(2^H * m) * (1 + epsAll)

          which implies that we have m, n, G and H (in range) and epsAll
          satisfying abs(epsAll) < 2^-124 such that 

           abs(5^G / 2^H - m / n) = abs(5^G / 2^H * epsAll) 
                                  <= 1 * abs(epsAll) 
                                  < 2^-124.

          This is a contradiction to the theorem stating that the worst-case 
	  for the inexact approximation problem "5^G / 2^H approximated by m / n"
          is 2^-117.... (TODO).

          Therefore we have the following implication:

           If m != 5^G * 2^(-H) * n 
           then 2^64 * righth + rightm != 2^64 * lefth + leftl.

          Together with (iv) this yields to 

           m is equal to 5^G * 2^(-H) * n if and only if 
            2^64 * righth + rightm is equal to 2^64 * lefth + leftl.

     So, with points (i) through (vi), we know that we can execute the following 
     algorithm:
     
     1st step: 

      If 2^64 * righth + rightm - 2^64 * lefth - leftl < 0 
      then return the appropriate boolean result for 2^H * m > 5^G * n.

     2nd step:

      If 2^64 * righth + rightm - 2^64 * lefth - leftl = 0 
      then return the appropriate boolean result for 2^H * m = 5^G * n.

     3rd step:

      Otherwise, we have 2^64 * righth + rightm - 2^64 * lefth - leftl > 0 
      then return the appropriate boolean result for 2^H * m < 5^G * n.
 
     We hence start by computing 

      2^64 * resh + resl = 2^64 * righth + rightm - 2^64 * lefth - leftl.

     As it is possible to show that

      -2^127 < -2^127 + 2^72 <= 2^64 * resh + resl <= 2^126 - 2^72 < 2^127 - 1

     we are sure that there will be no borrow-out resp. wrap-around 
     for this two-word substraction.

  */
  resl = rightm - leftl; /* Account for borrow afterwards. */
  resh = righth - lefth; 

  /* Account for borrow between low-order and high-order words. */
  if (leftl > rightm) resh--;

  /* Get a signed-variable representation of resh. 

     Remark: on the register level, this operation is a no-op.  It's
     just needed to convince the compiler to compare with zero in the
     next step.

  */
  reshsigned = *((int64_t *) &resh);

  /* 2^64 * resh + resl is negative iff reshsigned is negative */
  if (reshsigned < 0ll) {
    /* In this case, we have 2^H * m > 5^G * n, i.e. abs(x) > abs(y).
       We know that sgn(x) = sgn(y).

       For a less-or-equal comparison x <= y, the result is true iff 
       x (and y) is negative.
    */

    return xIsNeg;
  }

  /* If we are here, we have 2^H * m <= 5^G * n, i.e. abs(x) <= abs(y). 
     We know that sgn(x) = sgn(y).   

     We continue by filtering the exact case, i.e. when x = y.
  */
  if ((resh == 0ull) && (resl == 0ull)) {
    /* In this case, we have 2^H * m = 5^G * n, i.e. abs(x) = abs(y).
       We know that sgn(x) = sgn(y). This means, we have x = y.

       For a less-or-equal comparison x <= y, the result is true.

    */

    return 1;
  } 

  /* If we are here, we have 2^H * m < 5^G * n, i.e. abs(x) < abs(y). 
     We know that sgn(x) = sgn(y).   

     For a less-or-equal comparison x <= y, the result is true iff 
     x (and y) is not negative.
  */
  
  return !xIsNeg;
}


/* An empty dummy function in this translation unit for timing purposes */
int emptyFunctionBinary64Decimal64(_binary64 x, _Decimal64 y) {
  return 1;
}
